"""
Search objects for representing search parameters.

We also provide methods for conveting these values
into something handlers can understand.
"""

from dataclasses import dataclass, asdict, field
from typing import Any, Optional
from urllib.parse import urlencode


@dataclass
class SearchParam:
    """
    SearchParam - Providing search options for efficient searching.

    We define the following values:

        * filter - Term to search for(i.e, 'Inventory Mods')
        * index - Page index of results to view
        * pageSize - Number of items to display per page
        * gameVersion - Game version to search under
        * sort - Sorting method to use

    If any of these parameters are unnecessary,
    then the handler can ignore them.
    We use camel case for the parameter names.

    Users should probably use the setter methods for 
    configuring this class!
    Users can also use the 'set' method to configure
    all search parameters in only one call.
    """

    searchFilter: Optional[str] = field(default=None) # Term to search for
    index: Optional[int] = field(default=None)  # Page of search results to view
    pageSize: Optional[int] = field(default=None)  # Number of items to display per page
    gameVersion: Optional[int] = field(default=None)  # Game version to use
    sort: Optional[int] = field(default=None) # Sort method to use

    # Some sort options defined here:

    FEATURED = 0
    POPULARITY = 1
    LAST_UPDATE = 2
    NAME = 3
    AUTHOR = 4
    TOTAL_DOWNLOADS = 5


    def asdict(self) -> dict:
        """
        Converts ourselves into a dictionary.

        :return: Dictionary of SearchParam values
        :rtype: dict
        """

        return asdict(self)


def url_convert(search: SearchParam, url: str='') -> str:
    """
    Converts the given search object into valid URL parameters.
    If the 'base_url' is provided,
    then we will append the converted values to the base_url.

    :param search: The SearchParam to convert
    :type search: SearchParam
    :param url: URL to build off of
    :type url: str
    :return: Converted SearchParameter value
    :rtype: str
    """

    # Get a dictionary of parameters:

    params = search.asdict()
    final = {}

    # Iterate over the parameters:

    for key in params:

        # Check to ensure value is not None:

        if params[key] is not None:

            # Add this to the final:

            final[key] = params[key]

    # Encode and return the values:

    return url + urlencode(final)



"""
Low level tools and components for network communication.

We offer components that can be used to get curseforge information from somewhere,
in this case HTTP get requests.
"""

import socket

from urllib.request import urlopen, Request
from urllib.parse import urlencode
from http.client import HTTPResponse
from typing import Optional


class BaseProtocol(object):

    """
    Base protocol implementation - All child protocols must inherit this class!

    A 'protocol' is a class that eases low level network communications.
    Protocols only handle the getting(And maybe transmission) of information.
    Protocols do not engage in any unnecessary content parsing, interpreting, or decoding!

    Protocols are designed to be modular, and can be 'attached' to implementations/backends.
    This is not something the user will have to do, handlers will auto-configure
    the protocol they are designed to use.

    We keep this implementation ambiguous, although we do define some behavior that all
    protocols MUST implement!
    """

    def __init__(self, host:str, port, timeout:int=60) -> None:
        
        self.timeout = timeout  # Timeout value for this object
        self.host = host # Hostname of the entity we are connected to
        self.port = port  # Port number of the entity we are connected to

        self.total_sent = 0  # Total bytes sent in our lifetime
        self.total_received = 0  # Total bytes received in our lifetime


class TCPProtocol(BaseProtocol):
    """
    TCPProtocol - Protocol object for data exchange via TCP sockets.

    We offer convince methods for working with TCP sockets.
    We also provide methods for generating valid metadata,
    which will provide the number of bytes transmitted/received,
    as well as the remote entity we are communicating with.
    """

    def __init__(self, host: str, port: int, timeout: int=60):

        super().__init__(host, port, timeout=timeout)

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.last = {'bytes': 0}

        self.sock.settimeout(self.timeout)
        self.sock.connect((self.host, self.port))

    def read(self, length: int) -> bytes:
        """
        Reads content from the TCP socket.
        We make sure to check the received data length
        to ensure that we read everything that we expect.

        :param length: Number of bytes to read
        :type length: int
        :return: Retrieved bytes 
        :rtype: bytes
        """

        byts = b''

        # Continuosly loop:

        while len(bytes) < length:

            to_read = length - len(byts)

            # Read as many bytes as we can:

            last = self.sock.recv(4096 if to_read > 4096 else to_read)

            byts = byts + last

        # Update the metadata:

        self.last['bytes'] = len(byts)

        # Return the bytes:

        return byts

    def write(self, byts: bytes):
        """
        Writes data over the TCP socket.

        :param byts: Bytes to write
        :type byts: bytes
        """

        # Send the bytes:

        self.sock.sendall(byts)

        # Update our metadata:

        self.last['bytes'] = len(byts)


class UDPProtocol(BaseProtocol):
    """
    UDPProtocol - Protocol object for data exchange via UDP sockets.

    We offer convenience methods for working with UDP sockets.
    Like TPCProtocol, we keep metadata that keeps track of data transferred,
    as well as the address of the last exchange.
    """

    def __init__(self, host: str, port, timeout: int):

        super().__init__(host, port, timeout=timeout)

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.last = {'bytes': 0,'addr':  None}

        self.sock.settimeout(self.timeout)

    def read(self) -> bytes:
        """
        Reads content over the UPD socket.

        :return: Bytes read
        :rtype: bytes
        """

        # Get the data:

        resp = self.sock.recvfrom(1024)

        # Set the metadata:

        self.last['bytes'] = len(resp[0])
        self.last['addr'] = resp[1]

        # Return it:

        return resp[0]

    def write(self, byts: bytes):
        """
        Writes data over the UPD socket.

        :param byts: Bytes to write
        :type byts: bytes
        """

        # Send the data:

        self.sock.sendto(byts, (self.host, self.port))

        # Set the metadata:

        self.last['bytes'] = len(byts)
        self.last['addr'] = (self.host, self.port)


class URLProtocol(BaseProtocol):

    """
    URLProtocol - Gets information via HTTP.

    We seek to ease the process of retriving information via HTTP requests.
    We not only facilitate the communication process,
    but we also provide other features such as user defined headers,
    URL genration, and we implement the protocol caching system.

    The host will be used to automatically build URLs if used.
    If you want to provide URLs manually, you can use lower level methods to do so.

    We raise the usual urllib exceptions if issues arise.
    """

    def __init__(self, host: str, timeout: int=60) -> None:

        super().__init__(host, 80, timeout=timeout)

        self.headers = {'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36'}  # Request headers to use
        self.extra = '/'  # Extra information to add before the path when building URLs
        self.proto_id = 'URLProtocol'  # Protocol ID
    
        self.meta = {} # MetaData from the last request

    def get_data(self, url: str, timeout: Optional[int]=None, data: Optional[dict]=None) -> bytes:
        """
        Gets and returns raw data from the given URL.

        By default, we get this data via a HTTP GET request,
        and return the raw bytes from this request.

        The user can optionally provide a dictionary of data,
        which will turn this call into a POST operation.

        We handle caching, if enabled, and will automatically 
        get and update values from said cache.

        :param url: URL to get data from
        :type url: str
        :param timeout: Timeout of the operation, default value if None is used
        :type timeout: Optional[dict], optional
        :param data: Data to use in the call, converting this to a POST operation
        :type data: dict
        :return: Raw string data from the given URL
        :rtype: str
        """

        # Get the respone object:

        req = self.low_get(url, timeout=timeout)

        # Create the metadata:

        self._create_meta(req)

        # Get the raw content:

        cont = req.read()

        # Finally, return the data:

        return cont

    def low_get(self, url: str, timeout: Optional[int]=None, heads: Optional[dict]=None) -> HTTPResponse:
        """
        Low-level get method.

        We get the necessary data from the server and return
        the corresponding HTTP object.

        If you want to work with HTTPResponse objects directly,
        (Like if you want to get information regarding return codes and other information),
        then this is the method you should use!

        :param url: URL to get data from
        :type url: str
        :param timeout: Timeout value, uses default value if None
        :type timeout: int, optional
        :param heads: Extra headers to include with our defaults
        :return: HTTPResponse object contaning response from server
        :rtype: HTTPResponse
        """

        # Create the request:

        req = self._request_build(url, heads=heads)

        # Get the HTTPResponse object:

        self.last = urlopen(req, timeout=timeout if not None else self.timeout)

        # Return the object:

        return self.last

    def url_build(self, path: str) -> str:
        """
        Builds and returns a URL using the given path.

        We combine the hostname of this protocol instance,
        and the given path to generate a valid url.

        :param path: Path to append onto the end of the hostname
        :type path: str
        :return: New URL to use
        :rtype: str
        """

        # Combine and return the new URL:

        return self.host + self.extra + path

    def make_meta(self) -> dict:
        """
        Gets and returns the metadata on the last made request.

        This is a good way to integrate connection stats into your CurseInstances!

        The meta information is structured like this:

            * headers - A list of (header, value) tuples
            * version - HTTP Protocol version used by server
            * url - URL of the resource retrieved
            * status - Status code returned by server
            * reason - Reason phase returned by server
    
        These values are returned in dictionary format.

        :return: Dictionary of metadata 
        :rtype: dict
        """

        # Return the created metadata:

        return self.meta

    def _create_meta(self, resp: HTTPResponse):
        """
        Creates valid metadata from a HTTPResponse

        :param: HTTPResponse object to create metadata from
        :type resp: HTTPResponse
        :return: Dictionary of metadata 
        :rtype: dict
        """

        # Create and set the metadata:

        self.meta = {'headers': self.last.getheaders(), 'version': self.last.version, 
                        'url': self.last.geturl(), 'status':self.last.status, 'reason': self.last.reason}

    def _request_build(self, url: str, data: Optional[dict]=None, heads: Optional[dict]=None) -> Request:
        """
        Builds an urllib request object using the given parameters.

        We point the request at the given url,
        add content headers, and add the given data.

        :param url: URL of the request
        :type url: str
        :param data: Data to add, None if there is no data
        :type data: dict
        :param heads: Extra headers to optionally add to the request
        :type heads: dict  
        :return: Request object of this request
        :rtype: Request
        """

        # Check if we should encode the data:

        encoded_data = None

        if data is not None:

            encoded_data = urlencode(data).encode()

        # Convert heads to empty dictionary if necessary: 

        if heads is None:

            heads = {}

        # Make and return the request:

        return Request(url, data=encoded_data, headers={**self.headers, **heads})



"""
Stores Handler maps to be used by cursepy.

We store these in a separate folder to resolve circular import dependencies!
Here is a list of the follwing maps:

    * DEFAULT_MAP - Default handler map
"""

from cursepy.handlers.forgesvc import svc_map


# Default cursepy handler map:

DEFAULT_MAP = (svc_map)



"""
cursepy excpetions and errors
"""


class CurseBaseException(BaseException):

    """
    CFABaseException - Base exception all CFA excaptions will inherit!

    This will NOT be raised by any CFA component!
    This exception can be used to identify the custom CFA exceptions.
    """

    pass


class ProtocolMismatch(CurseBaseException):
    """
    Exception raised when a protocol mismatch has occurred.

    A 'protocol mismatch' is when a handler attempts to assign
    a protocol object that does not match the protocol object 
    currently assigned to the handler name.
    """

    pass


class HandlerRaise(CurseBaseException):
    """
    Exception raised when the 'RaiseHandler' is called.

    This exception will not be raised anywhere else!
    """

    pass



"""
ForgeSVC - Handlers for getting info via ForgeSVC.net
"""

from __future__ import annotations

import json
from typing import Any, Tuple

from cursepy.classes.search import SearchParam, url_convert
from cursepy.handlers.base import URLHandler
from cursepy.classes import base


class BaseSVCHandler(URLHandler):
    """
    BaseSVCHandler - Base handler all classes must inherit!

    We define some useful functionality that is relevant to child handlers,
    specifically setting our base URL, and using JSON for pre-processing.
    """

    def __init__(self):

        super().__init__('ForgeSVC', 'https://addons-ecs.forgesvc.net', extra='/api/v2/', path='')

        self.raw: Any  # Raw data before formatting and post-processing

    def pre_process(self, data: bytes) -> dict:
        """
        Decodes the bytes from the protocol object
        into JSON format.

        :param data: Bytes from the protocol object
        :type data: bytes
        :return: Dictionary representing raw data
        :rtype: dict
        """

        # Decode the data:

        self.raw = json.loads(data)
    
        # Return the raw data:

        return self.raw

    def post_process(self, data: Any, raw: dict=None, meta:dict=None) -> Any:
        """
        Post-process the packet.

        We add the raw data from the request,
        and we add on the metadata generated by URLHandler.

        The user can optionally provide raw data and metadata to attach to 
        the instance. This will overide any instances of the data we currently have.

        :param data: CurseInstance to be post-processed
        :type data: base.BaseCurseInstance
        :param raw: Raw data to be added
        :type raw: dict
        :param meta: Metadata to be added
        :type meta: dict
        :return: Processed curse instance
        :rtype: BaseCurseInstance
        """

        # Add the raw data to the instance:

        data.raw = self.raw if raw is None else raw

        # Add the metadata:

        data.meta = self.make_meta() if meta is None else meta

        # Return the final packet:

        return data


class SVCListGame(BaseSVCHandler):
    """
    Handler for getting a list of games.
    """

    ID: int = 1

    def build_url(self) -> str:
        """
        Builds a valid URL for getting game info.

        :return: URL for list of games
        :rtype: str
        """

        return self.proto.url_build('game')

    @staticmethod
    def format(data: dict) -> Tuple[base.CurseGame, ...]:
        
        # Iterate over the games:

        final = []

        for game in data:

            # Create the game and append it:

            final.append(SVCGame.format(game))

        # Return the final game:

        return tuple(final)


class SVCGame(BaseSVCHandler):

    """
    Handler for getting and representing a curseforge game,
    whatever that may be.

    We convert the incoming data to a CurseInstance,
    as well as the function for creating a the valid URL.
    """

    ID: int = 2

    def build_url(self, game_id: int) -> str:
        """
        Returns a URL to fetch game information

        :param game_id: ID of the game to search for
        :type game_id: int
        :return: URL of the game info
        :rtype: str
        """

        return self.proto.url_build('game/{}'.format(game_id))

    @staticmethod
    def format(data: dict) -> base.CurseGame:
        """
        Formats the given data into a CurseGame.

        :param data: Decoded data to convert
        :type data: dict
        :return: CurseGame instance representing the game
        :rtype: CurseGame
        """
        
        # Create list of game categories:

        final = []

        for cat in data['categorySections']:

            # Add the root ID's to the game:

            final.append(cat['gameCategoryId'])

        # Create CurseGame:

        return base.CurseGame(data['name'], data['slug'], data['id'], data['supportsAddons'], tuple(final))


class SVCListCategory(BaseSVCHandler):
    """
    Handler for getting all valid categories on CurseForge.

    We convert the given data into a tuple of CurseCategory instacnes.
    """

    def build_url(self) -> str:
        """
        # Returns a valid URL for fetching category information

        :return: URL for getting all catagories
        :rtype: str
        """

        return self.proto.url_build('category')

    def format(self, data: dict) -> Tuple[base.CurseCategory, ...]:
        """
        Formats the given data into a tuple of CurseCatagories

        :param data: Data to be formatted
        :type data: dict
        :return: Tuple of CurseCatagories
        :rtype: Tuple[CurseCatagories, ...]
        """

        # Iterate over the catagories:

        final = []

        for cat in data:

            # Create the category:

            final.append(SVCCategory.format(cat))

        # Return the final tuple:

        return tuple(final)


class SVCCategory(BaseSVCHandler):
    """
    Handler for getting and representing a curseforge category via ForgeSVC.
    """

    ID: int = 4

    def build_url(self, category_id: int) -> str:
        """
        Returns a valid URL to fetch category information

        :param category_id: ID of the category
        :type category_id: int
        :return: URL of the category info
        :rtype: str
        """

        return self.proto.url_build('category/{}'.format(category_id))

    @staticmethod
    def format(data: dict) -> base.CurseCategory:
        """
        Formats the decoded data into a CurseCategory instance

        :param data: Data to be formatted
        :type data: dict
        :return: CurseCategory instance
        :rtype: base.CurseCategory
        """

        return base.CurseCategory(data['id'], data['gameId'], data['name'], data['rootGameCategoryId'], 
                                  data['parentGameCategoryId'], data['avatarUrl'], data['dateModified'])


class SVCSubCategory(BaseSVCHandler):
    """
    Handler for getting sub-categories
    for a given category.
    """

    ID: int = 5

    def build_url(self, category_id) -> str:
        """
        Builds a valid URL for getting a list of sub-categories.

        :param category_id: ID of the category
        :type category_id: int
        :return: URL for list of sub-categories
        :rtype: str
        """

        return self.proto.url_build('category/section/{}'.format(category_id))

    @staticmethod
    def format(data: dict) -> Tuple[base.CurseCategory, ...]:
        """
        Formats the decoded data into a CurseCategory instance

        :param data: Data to be formatted
        :type data: dict
        :return: Tuple of CurseCategory instances
        :rtype: Tuple[CurseCategory, ...]
        """

        # Iterate over the catagories:

        final = []

        for cat in data:

            # Format the data:

            final.append(SVCCategory.format(cat))

        # Return the final tuple:

        return tuple(final)


class SVCAddon(BaseSVCHandler):
    """
    Handler for getting addon information.
    """

    ID: int = 6

    def build_url(self, addon_id: int) -> str:
        """
        Builds a valid URL for getting addon information.

        :param addon_id: ID of the addon
        :type addon_id: int
        :return: URL for getting addon info
        :rtype: str
        """

        return self.proto.url_build('addon/{}'.format(addon_id))

    @staticmethod
    def format(data: dict) -> base.CurseAddon:
        """
        Formats decoded data into CurseAddon instances

        :param data: Data to be formatted
        :type data: dict
        :return: CurseAddon instance representing the addon
        :rtype: base.CurseAddon
        """
        
        # Convert the authors:

        authors = []

        # Iterate over the authors:

        for auth in data['authors']:

            # Create and add the author:

            authors.append(base.CurseAuthor(auth['id'], auth['name'], auth['url']))

        # Convert the attachments:

        attach = []

        for attachment in data['attachments']:

            # Convert and add the attachment:

            attach.append(base.CurseAttachment(attachment['title'], attachment['id'], attachment['thumbnailUrl'],
            attachment['url'], attachment['isDefault'], attachment['projectId'], attachment['description']))

        # Create the instance:

        return base.CurseAddon(data['name'], data['slug'], data['summary'], data['websiteUrl'],
                               data['primaryLanguage'], data['dateCreated'], data['dateModified'], data['dateReleased'],
                               data['id'], data['downloadCount'], data['gameId'], data['isAvailable'], data['isExperiemental'],
                               tuple(authors), tuple(attach), data['primaryCategoryId'], data['isFeatured'], data['popularityScore'],
                               data['gamePopularityRank'], data['gameName'])


class SVCSearch(BaseSVCHandler):
    """
    Handler for searching for addon information.
    """

    def build_url(self, game_id: int, section_id: int, search: SearchParam) -> str:
        """
        Generates a valid URL with the given search parameter.

        :return: Valid URL for searching
        :rtype: str
        """

        # Create and return the URL:

        return self.proto.url_build(url_convert(search, url='addon/search?gameId={}&sectionId={}&'.format(game_id, section_id)))

    def format(self, data: dict) -> Tuple[base.CurseAddon, ...]:
        """
        Formats the given search results.

        :param data: Data to be formatted
        :type data: dict
        :return: Tuple of CurseAddon instances
        :rtype: Tuple[base.CurseAddon, ...]
        """

        # Iterate over the instances:

        final = []

        for addon in data:

            # Create the instance:

            final.append(SVCAddon.format(addon))

        # Return the final tuple:

        return tuple(final)


class SVCAddonDescription(BaseSVCHandler):
    """
    Handler for getting an addon description.
    """

    ID: int = 8

    def build_url(self, addon_id: int) -> str:
        """
        Builds a valid URL for getting addon descriptions.

        :param addon_id: Addon ID
        :type addon_id: int
        :return: URL for addon descriptions
        :rtype: str
        """
        
        return self.proto.url_build('addon/{}/description'.format(addon_id))

    def pre_process(self, data: bytes) -> str:
        """
        We do NOT decode the info via JSON,
        as ForgeSVC gives us HTML strings.

        :param data: Data to be converted into string
        :type data: bytes
        :return: String representing the HTML data
        :rtype: str
        """

        # Set the raw data:

        self.raw = data

        return data.decode()

    @staticmethod
    def format(data: str) -> base.CurseDescription:
        """
        Formats decoded data into a CurseDescription

        :param data: Data to be decoded
        :type data: dict
        :return: CurseDescription representing the addon description
        :rtype: base.CurseDescription
        """
        
        return base.CurseDescription(data)


class SVCAddonFiles(BaseSVCHandler):
    """
    Handler for getting all files for an addon
    """

    ID: int = 9

    def build_url(self, addon_id: int) -> str:
        """
        Builds a valid URL for getting a list of all files associated with an addon

        :param addon_id: ID of the addon
        :type addon_id: int
        :return: URL for getting file info
        :rtype: str
        """

        return self.proto.url_build('addon/{}/files'.format(addon_id))

    def format(self, data: dict) -> Tuple[base.CurseFile, ...]:
        """
        Formats the decoded data into a tuple of CurseFiles

        :param data: Data to be formatted
        :type data: dict
        :return: Tuple of curse files
        :rtype: Tuple[base.CurseFile, ...]
        """

        # Extract the addon ID from the URL:

        id = int(self.url.split('/')[6])

        # Iterate over the files:

        final = []

        for file in data:

            # Convert the file:

            final.append(SVCFile.low_format(file, id))

        # Return the data:

        return tuple(final)


class SVCFile(BaseSVCHandler):
    """
    Handler for getting addon files
    """

    ID: int = 10

    def build_url(self, addon_id: int, file_id: int) -> str:
        """
        Returns a valid URL for getting an addon file

        :param addon_id: ID of the addon
        :type addon_id: int
        :param file_id: ID of the file
        :type file_id: int
        :return: URL for addon file
        :rtype: str
        """

        return self.proto.url_build('addon/{}/file/{}'.format(addon_id, file_id))

    def format(self, data: dict) -> base.CurseFile:
        """
        Formats the given data into a CurseFile instance

        :param data: Data to be formatted
        :type data: dict
        :return: CurseFile instance
        :rtype: CurseFile
        """

        # Get the addon ID:

        id = int(self.url.split('/')[6])

        # Call the low-level format method:

        return self.low_format(data, id)

    @staticmethod
    def low_format(data: dict, addon_id: int) -> base.CurseFile:
        """
        Low-level format method.

        We are static to allow other classes to use us to format data.
        The actual 'format()' method will extract the addon ID from the URL
        and pass it along to us.

        :param data: Data to be formatted
        :type data: dict
        :param addon_id: ID of the addon
        :type addon_id: int
        :return: CurseFile instance
        :rtype: base.CurseFile
        """

        # Get the dependencies ID's:

        final = []

        for depen in data['dependencies']:

            # Add the dependency ID's:

            final.append(depen['addonId'])

        return base.CurseFile(data['id'], addon_id, data['displayName'], data['fileName'], data['fileDate'], 
        data['downloadUrl'], data['fileLength'], tuple(data['gameVersion']), tuple(final))


class SVCFileDescription(BaseSVCHandler):
    """
    Handler for getting the description of a file
    """

    ID: int = 11

    def build_url(self, addon_id: int, file_id: int) -> str:
        """
        Returns a valid URL for getting file description

        :param addon_id: ID of the addon
        :type addon_id: int
        :param file_id: ID of the file
        :type file_id: int
        :return: URL for getting file description
        :rtype: str
        """

        return self.proto.url_build('addon/{}/file/{}/changelog'.format(addon_id, file_id))

    def pre_process(self, data: bytes) -> str:
        """
        We do NOT decode the data via JSON,
        as we are working with HTML.

        :param data: Data to be decoded
        :type data: bytes
        :return: String representing the description
        :rtype: str
        """

        self.raw = data

        return data.decode()

    def format(self, data: str) -> base.CurseDescription:
        """
        Formats the given data into a CurseDescription instance

        :param data: Data to format
        :type data: str
        :return: CurseDescription representing the data
        :rtype: CurseDescription
        """

        return base.CurseDescription(data)


# Set the handler map:

svc_map = (SVCListGame(),
    SVCGame(),
    SVCListCategory(),
    SVCCategory(),
    SVCSubCategory(),
    SVCAddon(),
    SVCSearch(),
    SVCAddonDescription(),
    SVCAddonFiles(),
    SVCFile(),
    SVCFileDescription())



"""
Formatters for use with CurseDescriptions.

We offer methods for normalising and formatting descriptions,
which are in HTML.

This can be as simple as striping HTML markups,
or as in-depth as loading said information into a custom class.
"""

import re

from typing import Any


class BaseFormat(object):
    """
    BaseFormat - Base class all formatters must inherit!

    We define some useful functionality for all formatters to implement.
    A 'formatter' is something that changes HTML content into something else.
    What the content gets converted to is up to the formatter!
    """

    def format(self, data: str) -> Any:
        """
        Formats the given string data.

        This is the entry point for the formatter!
        If formatting is enabled,
        then this method will be called with the given string to be formatted.

        :param data: Data to be formatted
        :type data: str
        :return: Formatted data, whatever that may be
        :rtype: Any
        :raises: NotImplementedError: Must be overridden in child class!
        """

        raise NotImplementedError("Must be implemented in child class!")


class NullFormatter(BaseFormat):
    """
    NullFormatter - As the name implies, does nothing!

    We simply return the data that we were provided with.
    """
 
    def format(self, data: str) -> str:
        """
        Returns the given data

        :param data: Data to format
        :type data: str
        :return: Given data
        :rtype: str
        """

        return data


class StripHTML(BaseFormat):
    """
    Strips all HTML tags and metadata using regular expressions.

    This is not a foolproof operation,
    and this could easily fail in many ways!
    """

    PATTERN = '<[^>]*>'

    def format(self, data: str) -> str:
        """
        Strips the given data of all HTML tags and metadata.

        :param data: Data to format
        :type data: str
        :return: Formatted string data
        :rtype: str
        """

        return str(re.sub(StripHTML.PATTERN, '', data))
 

class BSFormatter(BaseFormat):
    """
    Loads the data into the Beautiful Soup parser for further processing.

    The user can specify the parser to use.
    We use python's built in HTML parser by default,
    although another can be selected using class parameters.
    """

    DEFAULT='html.parser'
    LXML='lxml'
    HTML5LIB='html5lib'

    def __init__(self, parser='html.parser'):
        
        self.parser = parser

    def format(self, data: str) -> Any:
        """
        Sends the given data to the Beautiful Soup parser.
        We use the selected parser.

        :param data: HTML Data to parse
        :type data: str
        :return: BeautifulSoup object
        :rtype: Any
        """

        import bs4

        return bs4.BeautifulSoup(data, self.parser)



"""
This file contains base components for handler development.

A 'handler' is something that 'handles' data from a specific source of a specific type.
Essentially, when data comes from a specific source,
the handler is the thing that gets and formatts that data into something that is easy to work with.
A handler may also fetch data from elsewhere to fill in the blanks.

In this file, we outline some of the functionality of a handler,
and layout certain components that it will need to play nicely with everything else.

Because handlers can do a large variety of things,
the specific implementation is left ambiguous.

By default, the HandlerCollection only works with the 'handle' method,
which will accept raw data from the service we are talking to.
There will be other components that the handler can overide if necessary.
"""

from typing import Any, Callable, Optional

from cursepy.classes import base
from cursepy.proto import BaseProtocol, URLProtocol
from cursepy.errors import ProtocolMismatch, HandlerRaise
from cursepy.classes.search import SearchParam
from cursepy.formatters import BaseFormat, NullFormatter


class BaseHandler(object):
    """
    BaseHandler - Child class all handlers must inherit!

    We offer some basic methods that a handler should inherit,
    as well as define the cappy Handler Framework(CHF).

    Essentially, the function makeup of the CHF is this:

        * make_proto - Should return a valid protocol object for this handler
        * proto_call - Call to the protocol of this handler, should return valid raw data
        * pre_process - Decodes data from the protocol before being send to the handler
        * format - Method called for formatting raw data from the protocol
        * post_process - Finalizes the formatted data(Adds metadata, raw data, ect.)
        * handle - Invokes the handling process for this handler

    This allows handlers to control many aspects of their implementation.
    With smart inheritance, it also removes a lot of repeated calls and instructions.
    These functions are called in the order defined above.
    Everything from getting information to formatting can be customized. 

    Handlers are identified by instance types.
    Instance types are simply an intiger that represents what this handler does.

    The handle method is the entry point for handlers.
    It will accept the arguments and information from the handler collection,
    and will work with them as necessary.
    The default handle method will call all componets in order,
    so it is a good idea to use 'super()' once you have configured your methods.
    You can configure the handle method to accept any arguments you like, 
    the HandlerCollection will pass all arguments that it receives to the 'handle' method,
    so you should use this as the primary way to accept arguments. 

    The handler ID will be stored under the ID parameter.
    Unidentified handlers will have an ID of -1.
    Handlers must be identified if they are to be used for handler processes!

    These handlers are designed to be customizable!
    The CHF is designed to be a good default implementation,
    that minimizes the amount of code developers have to write for handler development.

    Don't like the CHF? No problem!
    Most of the functions defined here are not necessary for handlers to work!
    The only functions that NEED to be defined are 'handle' and 'make_proto'.
    This means you can define your own methods and frameworks by creating a custom 'handle' method,
    that can call/do anything you would like it too.
    """

    ID: int = -1

    def __init__(self, name: str='') -> None:
        
        self.hand_collection: HandlerCollection  # Handlers instance we are apart of
        self.name = name  # Name of this handler, used to identify like-minded handlers

        self.proto: BaseProtocol  # Underlying protocol object in use

    def start(self):
        """
        Function called when this handler is attached to a HandlerCollection.

        The user can put any relevant code here
        that would be appropriate to run at loadtime.
        """

        pass

    def stop(self):
        """
        Function called when this handler is removed from a HandlerCollection.

        Again, the user can put any relevant code here
        that would be appropriate to run at stoptime.
        """

        pass

    def format(self, data: Any):
        """
        Converts the given raw data into something,
        usually a CurseInstance.

        In here, we format and organise the given data
        into something that is usefull to the user.

        This formatter can do other things if necessary,
        and they don't necessarily have to be within the scope of formatting!
        However, it is recommended to use the 'handle' method for other use cases.

        :param data: Data to be formatted
        :type data: Any
        """

        # Return the data given to us

        return data

    def proto_call(self) -> Any:
        """
        Calls the underlying protocol instance of this handler.

        This function only calls the protocol instance!
        We do not decode or interpret this information,
        we only get it and return.

        :return: Raw data from the protocol
        :rtype: Any
        """

        pass

    def pre_process(self, data: Any) -> Any:
        """
        We pre-process the raw information for the protocol,
        and convert it into something the handler can understand.

        :param data: Data to decode
        :type data: Any
        """

        pass

    def post_process(self, data: Any) -> Any:
        """
        Finalizes the packet after it is created.

        This can do many things, attach raw data,
        metadata, anything the developer thinks is necessary.

        :param data: Data packet to be finalized
        :type data: Any
        :return: Finalized packet
        :rtype: Any
        """

        pass

    def handle(self, *args, **kwargs) -> Any:
        """
        Default handle method.

        This is the default handle method,
        meaning that we call all components in order.

        We pass all arguments to 'proto_call',
        as this is the component that will need these arguments to fetch information.

        We also check to see if we are working with a tuple of instances.
        If this is the case, then we iterate over them and post-process each instance.
        """

        # First, call the proto_get method:

        data_bytes = self.proto_call(*args, **kwargs)

        # Next, call the pre-process method:

        raw_data = self.pre_process(data_bytes)

        # Format the data:

        data = self.format(raw_data)

        # Check if we are working with tuples:

        if type(data) == tuple:

            # Iterate over the data:

            final = []

            for item in data:

                # Process this item:

                post = self.post_process(item)

                # Attach the HandlerCollection:

                post.hans = self.hand_collection

                # Attach to the tuple:

                final.append(post)

            data_final = tuple(final)

        else:

            # Post-process the data:

            data_final = self.post_process(data)

            # Attach ourselves:

            data_final.hands = self.hand_collection

        # Return the final data:

        return data_final

    def make_proto(self) -> Any:
        """
        Function called when we need a valid protocol object.

        This usually occurs when a 'get_proto' check determines that no protocol exists for this handler.
        If this is the case, then this function should return a valid protocol object for this handler
        and other handlers of this same name.
        """

        return None

    def get_proto(self) -> BaseProtocol:
        """
        Gets the given protocol object from the HandlerCollection.

        If a protocol of our name is present,
        then we will get it for our continued use.
        If not, then we will create one and register it to our name.

        If a protocol of our name is present,
        and it does not match the type of the protocol we are attempting to assign,
        then a 'ProtocolMismatch' exception will be raised.

        :param proto: Protocol object type to fetch
        :type proto: BaseProtocol
        :return: Protocol instance that we will use
        :rtype: BaseProtocol
        :raises ProtocolMismatch: Raised when a protocol mismatch occurs
        """

        # Check if our name is present:

        temp_proto = self.make_proto()

        if self.name in self.hand_collection.proto_map:

            # Let's make sure our protocol is of our type:

            target = self.hand_collection.proto_map[self.name]

            if isinstance(target, type(temp_proto)):

                # Correct instance, lets return:

                return target

            # Invalid handler, let's do someting!

            raise ProtocolMismatch("Protocol is of type: '{}', must be of type: '{}'!".format(type(temp_proto), type(target)))

        # Not present, let's make our own:

        self.hand_collection.proto_map[self.name] = temp_proto

        return self.hand_collection.proto_map[self.name]


class URLHandler(BaseHandler):
    """
    URLHandler - Parent class for handlers using URLProtocol.

    We offer some ease of use features,
    such as automatically registering our protocol as URLProtocol,
    and simplifying the process for getting information from remote resources.
    We use the given information to automatically build URLs to get data from.

    We also offer ways to get the HTTPResponse of the last made request,
    and offer an easy way to generate metadata for curse instances.
    """

    def __init__(self, name: str, host: str, extra: str='/', path: str=''):

        super().__init__(name=name)

        self.host = host  # Hostname of the remote entity
        self.extra = extra  # Extra paths at the end of the hostname
        self.path = path  # Path to the correct data, appended after extra

        self.url = ''  # URL to use for the next request, can and SHOULD be overridden!

        self.proto: URLProtocol

    def make_proto(self) -> URLProtocol:
        """
        Creates and returns a valid URLProtocol instance.

        We automatically add the extra path information
        to the protocol instance.

        :return: Valid URLProtocol for this handler
        :rtype: URLProtocol
        """

        # Create the URL protocol

        url = URLProtocol(self.host)

        # Add the extra path:

        url.extra = self.extra

        # Return the final URLProtocol:

        return url

    def make_meta(self) -> dict:
        """
        Makes valid metadata about the request and connection
        and returns it.

        This is a good way to integrate connection stats into your CurseInstances!

        The meta information is structured like this:

            * headers - A list of (header, value) tuples
            * version - HTTP Protocol version used by server
            * url - URL of the resource retrieved
            * status - Status code returned by server
            * reason - Reason phase returned by server
    
        These values are returned in dictionary format.
        """

        # Get and return the metadata:

        return self.proto.make_meta()

    def proto_call(self, *args) -> bytes:
        """
        Creates a valid URL using 'build_url()',
        and then passes this value to the protocol instance.
        We pass all arguments to the 'build_url()' method.

        We return the raw data from the HTTP request for further processing.
        We also save the HTTPResponse for future reference,
        specifically to generate metadata fro curse instances.

        :return: Raw data from the protocol object
        :rtype: bytes
        """

        # Get and return the data:

        return self.proto.get_data(self.url)

    def build_url(self, *args) -> str:
        """
        Method called by the 'proto_call()' method when a URL needs to be built.

        This is useful for generating custom URL's that may change with diffrent requests.
        By default, we simply return the result of URLProtocol's build_url() method,
        with the path parameter as the argument.

        :return: Build URL to make our request to
        :rtype: str
        """

        # Build our URL with the path:

        return self.proto.url_build(self.path)

    def handle(self, *args) -> Any:
        """
        We do the same as the BaseHandler's handle method,
        except we call the 'build_url' method to generate a valid URL.

        We pass all arguments to the 'build_url()' method.

        :return: CursesInstance generated from this handler
        :rtype: BaseCurseInstance
        """

        # Generate the URL:

        self.url = self.build_url(*args)

        # Pass it along to the parent handler:

        return super().handle()


class NullHandler(BaseHandler):
    """
    NullHandler - Does nothing

    As in the name, all information given to us will be ignored,
    and we will simply return 'None'.

    Great if you want to disable a certain feature.

    The only thing that needs to be specified is what ID
    we are being registered to.
    This is optional, as you can manually provide your own ID in the HandlerCollection.

    We are used as a default handler that is attached to ID's
    that have no handlers associated with them.
    """

    def __init__(self, name: str='Nullhandle', id: int=0) -> None:

        super().__init__(name=name)

        ID: int = id

    def handle(self, *args, **kwargs) -> None:
        """
        As we are a NullHandler, we simply return None.

        :return: None
        :rtype: None
        """

        return None


class RaiseHandler(BaseHandler):
    """
    Raises an exception when the 'handle' method is called.

    This is very similar to the NullHandler,
    but we go a step further by raising an exception.

    This can be used as a more explicit, forceful way
    of disabling a feature, compared to just doing nothing.

    You can optionally change the name of the handler,
    as well as provide a custom ID. 
    """

    def __init__(self, name: str='RaiseHandler', id: int=0) -> None:

        super().__init__(name=name)

        ID: int = id

    def handle(self, *args, **kwargs):
        """
        Raise a 'HandlerRaise' excpetion.

        :raises: HandlerRaise: Always
        """

        raise HandlerRaise()


class HandlerCollection(object):
    """
    HandlerCollection - Manages and works with handlers! 

    This allows wrappers and backends to easily manage
    and work with handlers.

    We normalize calling, adding, and removing handlers.
    This allows for handler autoconfiguration,
    and other time saving features.

    We also provide a collective space for handlers
    to share protocol objects between like-minded handlers.

    Only one handler can be registered to a given event!
    These handlers can be of diffrent types and come from diffrent sources,
    but only one can be registered to a given event!

    We define some entry points for calling handlers.
    These functions will be called by the users actually using these handlers.
    We define some arguments that should be defined,
    as well as what we expect the functions to return.

    Users can optionally attach callbacks to certain events.
    These callbacks should be a callable, which will be called after 
    the handle event is complete.
    Callbacks should accept at least one argument,
    which is the data the handler is returned.
    Any other arguments can be optionally provided.

    You can utilise wrappers to change what functions expect and return.
    Be aware, that default dataclasses will call the functions defined here
    with the given arguments!

    Sub-classing us is recommended and encouraged!
    This allows developers to easily automate actions,
    and make interacting with certain games much easier.

    By default, we call the 'load_default()' method upon instantiation.
    Wrappers should probably overload that method!
    If this is not something you want, you can pass False to the 'load_default'
    parameter.
    """

    LIST_GAMES = 0  # Gets a list of all valid games
    GAME = 1  # Get information on a specific game
    LIST_CATEGORY = 2  # Gets a list of all valid categories
    CATEGORY = 3  # Get information on a specific category
    SUB_CATEGORY = 4  # Get all sub-categories for the given category
    ADDON = 5  # Get information on a specific addon
    ADDON_SEARCH = 6  # Searches the game for addons
    ADDON_DESC = 7 # Get description for a specific addon
    ADDON_LIST_FILE = 8  # Gets a tuple of all files associated with an addon
    ADDON_FILE = 9  # Get information on a specific file for an addon
    FILE_DESCRIPTION = 10  # Description of a file
    
    DEFAULT_MAP: tuple = () # Default handler map

    def __init__(self, load_default=True):
        
        self.handlers = {}  # Dictionary of handler objects
        self.proto_map = {}  # Maps handler names to protocol objects
        self.callbacks = {}  # List of callbacks to run
        self.formatter = NullFormatter()  # Default formatter to attach to CurseDesciprion

        # Create a good starting state:

        self.reset()

        # Add the default handlers, if applicable:

        if load_default:

            self.load_default()

    def reset(self):
        """
        Resets the HandlerCollection back to it's original state.

        This clears all handlers and replaces them with NullHandlers.
        We also remove all affiliated protocol instances.

        Be warned, if not referenced,
        the removed handlers and all information on them may be erased!
        """

        # Remove all handlers:

        self.handlers.clear()

        # Create null handlers for each INST_ID:

        for num in range(11):

            # Create the NullHandler for this ID:

            self.add_handler(NullHandler(), id=num)

        # Remove the protocol objects:

        self.proto_map.clear()

        # Remove the callbacks:

        self.callbacks = {}

        # Remove the default formatter:

        self.formatter = NullFormatter()

    def add_handler(self, hand: BaseHandler, id: Optional[int]=None):
        """
        Adds the given handler to the HandlerCollection.

        By default, we pull the instance ID out of thew handler.
        However, the developer can manually provide an instance ID to register the handler to.
        If this is the case, then the ID on the handler will be ignored.

        We use 'remove_handler()' to remove the handler at this location.

        The handler to be added MUST inherit BaseHandler!
        If it does not, then a ValueError will be raised.

        :param hand: Handler to register
        :type hand: BaseHandler
        :param id: ID to register the handler to, None to use ID on given handler
        :type id: int
        :raises: TypeError: If the handler does not inherit BaseHandler
        """

        # Check if the handler is valid:

        if not isinstance(hand, BaseHandler):

            # Handler is not valid!

            raise TypeError("Handler does not inherit BaseHandler!")

        # Remove the handler at this position:

        self.remove_handler(hand.ID if id is None else id)

        # Add the handler:

        self.handlers[hand.ID if id is None else id] = hand

        # Attach ourself to the handler:

        hand.hand_collection = self

        # Check if a new protocol object is necessary:

        hand.proto = hand.get_proto()

        # Start the handler:

        hand.start()

    def remove_handler(self, id: int):
        """
        Removes the handler at the given ID.

        We replace the handler with a NullHandler
        to ensure that something is working with data no matter what.

        We will automatically stop the handler at this location.

        :param id: ID of the handler to remove
        :type id: int
        """

        # Check if a handler is even present:

        if id not in self.handlers.keys():

            # No handler present, just return:

            return

        # Stop the handler at the ID:

        self.handlers[id].stop()

        # Replace the handler with NullHandler:

        self.handlers[id] = NullHandler(id=id)

    def load_handlers(self, mapper: Any):
        """
        Loads handlers from the given iterable.

        This function allows multiple handlers to be loaded at once,
        and in a certain priority.

        We use the iterables to determine the priority and ID of the handler.
        During the load, if the handler at the given ID is NOT a NullHandler,
        then it will be skipped.
        This allows for handler priorities to be set.

        The 'root' iterable should contain sub-iterables
        that map handlers to ID's.
        The lower the index, or key, of the handler map,
        the higher priority they will have.
        This allows lower priority handlers to 'fill in the blanks'
        of there higher priority counterparts.

        The iterables can be lists or tuples,
        (Tuples are recommended for memory and speed reasons,
        and are the iterable method used for built in handlers).
        You can also use a dictionary to denote ordering.
        The keys for the dictionary should be an intiger.
        Higher integers mean they are lower priority.
        
        If the dictionary is used as a handler map,
        then the key should be the ID that the handler is going to
        associate with.

        Here is an example of an iterable for loading handlers:

        .. code-block::

            (
                (
                    Hand1(),
                    Hand1(),
                    Hand1()
                ),
                {
                    0: Hand2(),
                    2: Hand2(),
                    4: Hand2()
                }
            )

        In this example, the handler map will look like this:

        * 0 - Hand1
        * 1 - Hand1
        * 2 - Hand1
        * 4 - Hand2

        Notice, that even though Hand2 was assigned to ID's 0 and 2,
        it was not used, as Hand1 had a higher priority.
        However, ID 4 had no handlers assigned to it,
        so Hand2 was allowed to associate with it.

        Here is another example, and instead of utilizing
        dictionaries for skipped indexes, we simply use a NullHandler as a placeholder:

        .. code-block::

            (
                (
                    Hand1(),
                    Hand1(),
                    NullHandler(),
                    Hand1()
                ),
                (
                    Hand2(),
                    NullHandler(),
                    Hand2(),
                )
            )

        In this example, we do not assign a handler for the 2nd index
        in the first priority handler map, and instead assign a NullHandler in it's place.
        Because NullHandlers are ALWAYS a lower priority,
        the lower priorities will overrite them, even though they have been specified.
        This allows users to skip ID's without having to use dictionaries.

        We use 'add_handler()' under the hood,
        and we use recursion to iterate over the handlers.

        :param loader: Iterable to load
        :type loader: Any
        """

        iterable = None

        # Dictionary check:

        if type(mapper) == dict:

            # Working with a dictionary, change the iter method:

            iterable = sorted(mapper.items(), key= lambda item: item(0))

        else:

            # Check if what we are working with is a valid iterable:

            try:

                # Check if we can iterate:

                iter(mapper)

                # Set our iterable:

                iterable = enumerate(mapper)

            except TypeError:

                # Not a valid iterable!

                raise ValueError("Not a valid iterable!")

        # Now, iterate over the content:

        for num, value in iterable:

            # Check if we are registering a handler:

            if isinstance(value, BaseHandler) and type(self.get_handler(num)) == NullHandler:

                # Got a valid handler!

                self.add_handler(value, num)

                continue

            # Something else, lets use some recurision!

            self.load_handlers(value)

    def load_default(self):
        """
        Loads the default handlers for this HandlerCollection.

        We are called at instantiation if load_default is True.
        It can also be called at any point during the lifecycle of the collection.

        This function should load a set of handlers that is relevant to the 
        operation of the HandlerCollection.

        By default,
        we raise a 'NotImplementedError'.
        It is up to wrappers to implement this function!

        It is probably best if this function used the 'load_handlers()'
        method to do all the dirty work.
        """

        raise NotImplementedError("Must be implemented in child class!")

    def get_handler(self, id: int) -> BaseHandler:
        """
        Returns the handler at the given ID.

        :param id: ID of the handler to get
        :type id: int
        :return: The handler in question
        :rtype: BaseHandler
        """

        return self.handlers[id]

    def bind_callback(self, call: Callable, id: int, *args, **kwargs) -> Callable:
        """
        Binds a callback function to the handler
        at the given ID.

        We can either be called the conventional way,
        or we can be used as a decorator.

        Your run function should take at least one parameter,
        which will be the CurseInstance associated with the handler.
        Other arguments can optionally be provided.

        We keep a reference of the callback
        and arguments in this class as a dictionary.
        You can alter this structure yourself,
        but it is recommended to use the higher level methods for this.

        Any extra arguments that are passed will be saved and passed 
        to the callback at a later time.

        Multiple callbacks can be registered to an event!
        The order that they are added will be the order that they will be called.

        The callback will be called after the data is retrieved, decoded, and formatted!

        :param call: Callback function to add to the handler
        :type call: function
        :param id: ID of the handler to bind the callback to
        :type id: int
        """

        # Setup the basic structure:

        if id not in self.callbacks.keys():

            # Set the value to a list:

            self.callbacks[id] = []

        # Add the callback and argument info to the structure:

        self.callbacks[id].append((call, args, kwargs))

        return call

    def clear_callback(self, id: int, call: Callable=None) -> int:
        """
        Clears the callback from the given event.

        If a callback is not specified,
        then ALL callbacks will be cleared.
        If a callback is specified, 
        then the callbacks that match the provided object will be deleted.

        :param id: ID to clear
        :type id: int
        :param call: Callback to remove, optional
        :type call: Callbale, None
        :return: Number of callbacks removed.
        :rtype: int
        """

        removed = 0

        # Check if the ID is present:

        if id not in self.callbacks.keys():

            # Key is NOT present, return

            return removed

        # Iterate over callbacks at the ID:

        for index, val in enumerate(self.callbacks[id]):

            # Check if the callback matches:

            if val[0] == call or call is None:

                # Remove the callback:

                self.callbacks[id].pop(index)
                
                removed += 1

        # Return the number of items removed:

        return removed

    def get_search(self) -> SearchParam:
        """
        Returns a SearchParam used for searching.

        :return: SearchParam object used for searching
        :rtype: SearchParam
        """

        # Return a SearchParam object:

        return SearchParam()

    def default_formatter(self, form: BaseFormat):
        """
        Registers a default formatter to this object.

        We automatically attach the default formatter to
        any CurseDesciprion objects we return.

        If the CurseDescrpition does not inherit
        BaseFormat, then we will raise a TypeError excpetion.

        :param form: Formatter to register
        :type form: BaseFormat
        :raises: TypeError: If formatter does not inherit BaseFormat!
        """

        # Check if the given object inherits 'BaseFormatter':

        if not isinstance(form, BaseFormat):

            raise TypeError("Formatter must inherit 'BaseFormat'!")

        # Save the formatter instance:

        self.formatter = form

    def handle(self, id: int, *args, **kwargs) -> Any:
        """
        Invokes the handle process at the handler with the given ID.

        As outlined earlier,
        this gets and formats data from an external location.

        We pass all arguments and keyword arguments
        to the handler at the given ID.

        We also attach ourselves to valid CurseInstances automatically!
        If the returned data does not inherit BaseCurseInstance,
        then we will not attempt to attach ourselves to it,
        and simply return it.

        :param id: ID of the handler to call
        :type id: int
        :return: The final class provided by the handler
        :rtype: BaseCurseInstance
        """

        # First, we get the handler in question:

        hand = self.handlers[id]

        # Invoke the 'handle()' method:

        inst = hand.handle(*args, **kwargs)

        # Check if we are working with a tuple:

        if type(inst) == tuple:

            # Iterate over the tuple:

            for pack in inst:

                # Check if we are working with a valid instance:

                self._format_object(pack)

        else:

            # Check if we are working with a valid instance:

            self._format_object(inst)

        # Check if there is a valid event:

        if id in self.callbacks.keys():

            # Run all callbacks associated with the event:

            for call in self.callbacks[id]:

                # Run the callback:

                call[0](inst, *call[1], **call[2])

        # Return the instance:

        return inst

    def _format_object(self, pack):
        """
        Adds ourselves to any valid CurseInstances.

        :param pack: CurseInstance to format
        :type pack: BaseFormat
        """

        if isinstance(pack, base.BaseCurseInstance):

            # Attach ourselves to the instance:

            pack.hands = self

            # Check if we are working with CurseDescription:

            if isinstance(pack, base.CurseDescription):

                # Attach our formatter:

                pack.attach_formatter(self.formatter)



"""
cursepy wrapper classes.

These classes are designed to work with specific games,
and make getting information about them much easier.
The class you probably want to work with is the CurseClient.
"""

from typing import Tuple

from cursepy.handlers.base import HandlerCollection
from cursepy.classes import base
from cursepy.classes.search import SearchParam
from cursepy.handlers.maps import DEFAULT_MAP


class CurseClient(HandlerCollection):
    """
    CurseClient - the client class most people will use!

    We implement the handler management and callback features outlined in the HandlerCollection.
    We do some extra things here and there to make life much easier for the end user. 

    We define some entry points for calling handlers.
    These functions will be called by the users actually using these handlers.
    We define some arguments that should be defined,
    as well as what we expect the functions to return.

    This class also provides some entry points into the loaded handlers.
    This provides a standardized way of interacting with handlers.
    """

    def load_default(self):
        """
        We simply load the cursepy default handler map.
        """
        
        self.load_handlers(DEFAULT_MAP)

    def games(self) -> Tuple[base.CurseGame]:
        """
        Returns a tuple of all games supported on curseforge.

        This call can be somewhat intensive,
        so it's a good idea to not call it often.
        Make a note of the relevant game's information!

        :raises NotImplementedError: Must be overridden in child class!
        :return: Tuple of CurseGame instances
        :rtype: Tuple[base.CurseGame]
        """

        return self.handle(0)

    def game(self, id: int) -> base.CurseGame:
        """
        Returns information on a specific game.

        You will need to provide the game ID.
        Game information will be returned in a CurseGame instance.

        :param id: ID of the game to get
        :type id: int
        :raises NotImplementedError: Must be overridden in child class!
        :return: CurseGame instance representing the game
        :rtype: base.CurseGame
        """

        return self.handle(1, id)

    def catagories(self) -> Tuple[base.CurseCategory, ...]:
        """
        Gets ALL valid categories on CurseForge.

        This call can get expensive, so call in moderation!

        :return: Tuple of CurseCategory instances
        :rtype: Tuple[base.CurseCategory, ...]
        """

        return self.handle(2)

    def category(self, category_id: int) -> base.CurseCategory:
        """
        Returns information on a category for a specific game.

        You will need to provide a category ID.

        :param game_id: ID of the game
        :type game_id: int
        :param category_id: ID of the category
        :type category_id: int
        :raises NotImplementedError: Must be overridden in child class!
        :return: CurseCategory instance representing the category
        :rtype: base.CurseCategory
        """

        return self.handle(3, category_id)

    def sub_category(self, category_id: int) -> Tuple[base.CurseCategory, ...]:
        """
        Gets the sub-categories of the given category.

        We return a tuple of CurseCategory instances
        representing the sub-categories

        :param category_id: ID of category to get sub-catagories for
        :type category_id: int
        :return: Tuple of CurseCategories representing the sub-categories
        :rtype: Tuple[base.CurseCategory, ...]
        """

        return self.handle(4, category_id)

    def iter_search(self, game_id: int, category_id: int, search: SearchParam=None) -> base.CurseAddon:
        """
        Iterates over all results from the search operation.

        We automatically bump the SearchParam by one after
        each call. We will keep yielding until
        we receive a tuple that has a length of zero,
        which at that point we will raise a 'StopIteration' exception.

        Because we only bump the index value,
        we will start on the index that was provided by the
        SearchParam object. For example,
        if the index starts at one, then we will return
        each addon from that page on.

        We use the 'search' method under the hood.

        This method is best used in for loops/situations
        where iterators are used!
        Because we are a generator,
        calling this method on it's own is not recommended!
        If you want a list of values,
        use the 'list()' method on us.

        :param game_id: Game ID to search under
        :type game_id: int
        :param category_id: Category ID to search under
        :type cat_id: int
        :param search: Search Parameter to use
        :type search: SearchParam
        :return: Each CurseAddon that returned during the search operation.
        :rtype: base.CurseAddon
        :raises: StopIteration: When we have run out of search content
        """

        # Check if we should generate a SearchParam:

        if search is None:

            # None provided, let's create one:

            search = self.get_search()

        while True:

            # Get the current page of results:

            results = self.search(game_id, category_id, search)

            # Check to see if the page is empty:

            if not results:

                # No results! Raise 'StopIteration'!

                raise StopIteration()

            # Iterate over the result:

            for item in results:

                # Yield the result:

                yield item

            # Bump the index and continue:

            search.index =+ 1

    def addon(self, addon_id: int) -> base.CurseAddon:
        """
        Gets information on a specific addon.

        You will need to provide an addon ID,
        which can be found by searching a game category.

        :param addon_id: Addon ID
        :type addon_id: int
        :raises NotImplementedError: Must be overridden in child class
        :return: CurseCategory instance representing the addon
        :rtype: base.CurseAddon
        """

        return self.handle(5, addon_id)

    def search(self, game_id: int, category_id: int, search: SearchParam=None) -> Tuple[base.CurseAddon, ...]:
        """
        Searches the given game and category for addons.

        The game_id and category_id parameters are required for searching,
        but the search parameter can optionally be provided to fine-tune to search.

        Each implementation has diffrent search backends,
        so it is important to pass the correct search instance!

        The wrappers will implement search checking,
        to ensure only the correct search parameters are used. 

        :param game_id: ID of the game to search under
        :type game_id: int
        :param category_id: Category to search under
        :type category_id: int
        :param search: Search options to fine tune the search, optional
        :type search: Any
        :return: Tuple of addons that matched the search parameters
        :rtype: Tuple[base.CurseAddon, ...]
        :raises: TypeError: If the search object fails the search check
        """ 

        # Check if we should generate a dummy SearchParam:

        if search is None:

            # Create dummy search param:

            search = SearchParam()

        return self.handle(6, game_id, category_id, search)

    def addon_description(self, addon_id: int) -> base.CurseDescription:
        """
        Gets the description of a specific addon.

        You will need to provide an addon ID,
        which can be found by searching a game category.

        :param addon_id: Addon ID
        :type addon_id: int
        :raises NotImplementedError: Must be overridden in child class!
        :return: CurseDescription instance representing the addon's description
        :rtype: base.CurseDescription
        """

        return self.handle(7, addon_id)

    def addon_files(self, addon_id: int) -> Tuple[base.CurseFile]:
        """
        Gets a list of files associated with the addon.

        You will need to provide an addon ID,
        which can be found by searching a game category.

        :param addon_id: Addon ID
        :type addon_id: int
        :raises NotImplementedError: Must be overridden in child class!
        :return: Tuple of CurseFile instances representing the addon's files
        :rtype: Tuple[base.CurseFile]
        """

        return self.handle(8, addon_id)

    def addon_file(self, addon_id: int, file_id:int) -> base.CurseFile:
        """
        Gets information on a specific file associated with an addon.

        You will need to provide an addon ID,
        as well as a file ID, which can be found by getting a list of all files

        :param addon_id: Addon ID
        :type addon_id: int
        :param file_id: File ID
        :type file_id: int
        :raises NotImplementedError: Must be overridden in child class!
        :return: CurseFile instance representing the addon file
        :rtype: base.CurseFile
        """

        return self.handle(9, addon_id, file_id)

    def file_description(self, addon_id: int, file_id: int) -> base.CurseDescription:
        """
        Gets a description of an addon file.

        You will need to provide a addon ID and a file ID.
        We return a CurseDesciprion object representing the description.

        :param addon_id: Addon ID
        :type addon_id: int
        :param file_id: File ID
        :type file_id: int
        :return: CurseDescription representing the description
        :rtype: base.CurseDescription
        """

        return self.handle(10, addon_id, file_id)


class MinecraftWrapper(CurseClient):
    """
    MinecraftWrap - Wrapper for working with Minecraft data!

    Minecraft is one of the most popular games on CurseForge.
    Because of this, we think it is appropriate to include a wrapper
    for solely working with Minecraft data.

    We offer a lot of convince methods,
    most notably are the methods for searching/getting addons in a specific category.
    This will automate the search operation,
    filling in the necessary info when necessary.
    This means that you as the user will not have to work(as much) with IDs! 
    """

    # Game ID:

    GAME_ID = 432

    # Category IDs:

    RESOURCE_PACKS = 12
    MODPACKS = 4471
    MODS = 6
    WORLDS = 17

    def search_resource_packs(self, search: SearchParam=None) -> Tuple[base.CurseAddon, ...]:
        """
        Searches the Resource Packs category for addons.
        We use the given SearchParam for the search operation.

        :param search: SearchParam to use, default to None
        :type search: SearchParam, optional
        :return: Tuple of CurseAddons
        :rtype: Tuple[base.CurseAddon, ...]
        """

        # Search the resource packs:

        return self.search(MinecraftWrapper.GAME_ID, MinecraftWrapper.RESOURCE_PACKS, search)

    def search_modpacks(self, search: SearchParam=None) -> Tuple[base.CurseAddon, ...]:
        """
        Searches the Modpacks category for addons.
        Again, we use the SearchParam for the search operation.

        :param search: SearchParameters to use, defaults to None
        :type search: SearchParam, optional
        :return: Tuple of CurseAddons
        :rtype: Tuple[base.CurseAddon, ...]
        """

        # Search the modpacks:

        return self.search(MinecraftWrapper.GAME_ID, MinecraftWrapper.MODPACKS, search)

    def search_mods(self, search: SearchParam=None) -> Tuple[base.CurseAddon, ...]:
        """
        Searches the Mods category for addons.
        Again, we use the SearchParam for the search operation.

        :param search: SearchParam to use, defaults to None
        :type search: SearchParam, optional
        :return: Tuple of CurseAddons
        :rtype: Tuple[base.CurseAddon]
        """

        # Search the mods:

        return self.search(MinecraftWrapper.GAME_ID, MinecraftWrapper.MODS, search)

    def search_worlds(self, search: SearchParam=None) -> Tuple[base.CurseAddon, ...]:
        """
        Searches the Worlds category for addons.
        Again, we use the SearchParam for the search operation.

        :param search: SearchParam to use, defaults to None
        :type search: SearchParam, optional
        :return: Tuple of CurseAddons, optional
        :rtype: Tuple[base.CurseAddon, ...]
        """

        # Search the worlds:

        return self.search(MinecraftWrapper.GAME_ID, MinecraftWrapper.WORLDS, search)



"""
General classes for representing curseforge information.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Optional, Tuple
from os.path import isdir, join

from cursepy.classes.search import SearchParam
from cursepy.formatters import BaseFormat, NullFormatter
from cursepy.proto import URLProtocol


@dataclass
class BaseCurseInstance(object):
 
    """
    BaseCurseInstance - Class all child instances must inherit!

    A curses instance is a class representing information from curseforge.
    This allows developers to work with this information in a connivent way.

    Some classes might even do operations on the data if needed,
    or might be able to fetch data themselves.

    We seek to standardize curseforge information,
    allowing developers to work with with only one packet type,
    regardless of backend/method/source of information.

    We also provide a field for supplying metadata about the instance.
    This can really be anything the backend thinks is helpful,
    although we recommend keeping it to connection details and statistics.

    Packets can also have the raw, unformatted data attached to them.
    This data is NOT standardized,
    meaning that the raw data will most likely be diffrent across diffrent handlers.
    Be aware, that handlers are under no obligation to attach raw data to the packet!

    We have the following parameters:

        * raw - Raw data from the backend(Does not have to be provided!)
        * meta - Meta data on this instance. See the backend in question for information on metadata
        * hands - HandlerCollection instance
    """

    raw: Any = field(init=False, repr=False, default=None)  # RAW packet data
    meta: Any = field(init=False, repr=False, default=None)  # Metadata on this packet
    hands: 'HandlerCollection' = field(init=False, repr=False, compare=False)  # Handler Collection instance


@dataclass
class BaseWriter(BaseCurseInstance):
    """
    Parent class that adds writing functionality to the instance.

    This class provides tools for writing content to external files,
    and offers entry points for child classes to fine-tune and configure these writes.
    """

    def low_write_bytes(self, data: bytes, path: str, append: bool=False) -> int:
        """
        Writes the given data to an external file in bytes.

        By default, we overwrite the file at the given path!
        If this is not something you want, you can pass 'True'
        to the 'append' value. 
        This will append the description to the end of the file provided.

        We use the internal 'open' function built into python.
        Standard exceptions associated with this function will be raised accordingly!

        :param data: Data to write to the file
        :type data: bytes
        :param path: Path to the file to write, must be a path-like-object
        :type path: str
        :param append: Value determining if we are appending to the file, defaults to False
        :type append: bool, optional
        :return: Number of bytes written
        :rtype: int
        """

        # Open the file-like object:

        file = open(path, 'ab' if append else 'wb')

        # Write the content:

        return file.write(bytes(data))

    def low_write_string(self, data: str, path: str, append: bool=False) -> int:
        """
        Similar to low_write_bytes, 
        except we write the content as a string.

        :param data: Data to write to file
        :type data: str
        :param path: Path to file to write
        :type path: str
        :param append: Value determining if we are appending to the file, defaults to False
        :type append: bool, optional
        :return: Number of bytes written
        :rtype: int
        """

        # Open the file-like object:

        file = open(path, 'a' if append else 'wb')

        # Write the content:

        return file.write(str(data))

    def write(self, path: str, append: bool=False) -> int:
        """
        Write function for this class.

        Child classes should implement this function!
        This allows them to fine-tune the write process automatically,
        and pass the content to be written without the user having to specify it.

        Child classes should use the 'low_write' function to preform the operation.

        :param path: Path to the file to write to
        :type path: str
        :param append: Value determining if we should append instead of overwrite, defaults to False
        :type append: bool, optional
        :raises NotImplementedError: Function should be overridden in child class!
        :return: Number of bytes written
        :rtype: int
        """

        raise NotImplementedError("Should be overloaded in child class!")


@dataclass
class BaseDownloader(BaseWriter):
    """
    Parent class for all instances that download information.

    We offer some helper methods for achieving this.
    We expect the protocol to be HTTP.
    """

    def low_download(self, url: str, path: str=None) -> bytes:
        """
        Downloads the given data and returns it as bytes.

        We use the 'URLProtocol' to get the information.

        We also offer to save the downloaded information
        to an external file.
        You can pass the path to the file using the 'path' parameter.
        If this value is none, then the data will not be saved.
        The data will be returned regardless of weather it is saved or not.

        :param url: URL of the content to download
        :type url: str
        :param path: Path to file to save downloaded data, optional
        :type path: str, optional
        :return: Bytes of the downloaded data
        :rtype: bytes
        """

        # Create the URL protocol:

        url_proto = URLProtocol(url)

        # Get the data:

        data = url_proto.get_data('')

        # Determine if we should write to a file:

        if path is not None:

            # Save the data to a file:

            self.low_write_bytes(data, path)

        # Finally, return the data:

        return data

    def download(self, path: str=None) -> bytes:
        """
        Download function for this class!

        Child classes should implement this function!
        This allows them to fine-tune the download operation automatically,
        and pass the necessary parmeters without the user having to specify them.

        Implementations should use the 'low_download' function for this operation.

        :param path: Path to file to save the downloaded information to, defaults to None
        :type path: str, optional
        :return: Downloaded bytes
        :rtype: bytes
        :raises: NotImplementedError: Must be overridden in child class!
        """

        raise NotImplementedError("Must be overridden in child class!")


@dataclass
class CurseAuthor(BaseCurseInstance):
    """
    CurseAuthor - Represents an author for a curses addon.

    An 'author' is an account that published a curses addon.

    This class only contains metadata on a author.
    The information in this class is not necessary for 
    getting addon information.
    Instead, it is extra info for those interested.

    Completely disregard this info if it is not relevant to you!

    We contain the following parameters:

        * id - ID of this author
        * name - Name of the author
        * url - URL to author's page
    """

    id: int
    name: str
    url: str

    INST_ID = 3


@dataclass
class CurseDescription(BaseWriter):

    """
    CurseDescription - Represents a decritpion of a addon.

    A 'description' is HTML code that the addon author can provide
    that describes something in detail.
    Addons and file descriptions fall into this category.

    Again, this description is in HTML,
    so some formatting may be necessary!

    We offer the ability to attach custom 'formatters' to this object.
    A 'formatter' will convert the HTML code into something.
    The implementation of formatters is left ambiguous,
    as users may have diffrent use cases.
    Check out the formatter documentation for more info.

    A formatter can be attached to us manually, 
    or it can automatically be added upon each call by the CursesAPI
    or wrapper.

    We also offer the ability to save the description to an external file
    for later usage.

    We contain the following parameters:

        * description - Raw, unformatted description
        * formatter - Formatter instance attached to this object 
    """

    description: str
    formatter: BaseFormat = field(default=NullFormatter(), init=False)

    INST_ID = 7

    def attach_formatter(self, form: BaseFormat):
        """
        Attaches the given formatter to this instance.

        We do a check to make sure that the formatter inherits 'BaseFormat'.

        :param form: Formatter to attach to this description
        :type form: BaseFormat
        :raises: TypeError: If the formatter does NOT inherit 'BaseFormat'
        """

        # Check of the formatter inherits 'BaseFormat':

        if not isinstance(form, BaseFormat):

            # Object does not inherit BaseFormat!

            raise TypeError("Formatter must inherit 'BaseFormat'!")

        # Attach the formatter to ourselves:

        self.formatter = form

    def format(self) -> Any:
        """
        Sends the description through the formatter,
        and returns the formatted content,
        whatever that may be.

        :return: Formatted data
        :rtype: Any
        """

        return self.formatter.format(self.description)

    def write(self, path: str, append: bool=False) -> int:
        """
        Writes the description to an external file.

        We use the CurseWriter 'low_write' method for this operation.

        We write the raw content, as formatters may not return writable components!

        :param path: Path to the file to write, must be a path-like-object
        :type path: str
        :param append: Value determining if we are appending to the file, defaults to False
        :type append: bool, optional
        :return: Number of bytes written
        :rtype: int
        """

        # Use the 'low_write' method:

        return self.low_write_string(self.description, path, append=append)

    def __str__(self) -> str:
        """
        Method called when this object's content is requested as a string.

        We return the raw data, as it may not return string content.

        :return: Formatted description content
        :rtype: str
        """

        return self.description


@dataclass
class CurseAttachment(BaseDownloader):
    """
    CurseAttachment - Represents an attachment.

    An 'attachment' is a piece of media,
    usually an image or thumbnail,
    that is shown on the addon page.

    These attachments can be the icon of the addon,
    as well as bonus images of the addon provided by the author.

    We contain useful metadata on said images,
    and we also offer tools to download and save these images.

    We contain the following parameters:

        * title - Title of the attachment
        * id - ID of the attachment
        * thumb_url - URL of the thumbnail image(If attachment is image, then thumbnail is smaller image)
        * url - URL of the attachment
        * is_thumbnail - Boolean determining if this attachment is the thumbnail of the addon
        * description - Description of this attachment
    """

    title: str
    id: int
    thumb_url: str
    url: str
    is_thumbnail: bool
    addon_id: int
    description: str

    INST_ID = 4

    def download(self, path: str=None) -> bytes:
        """
        Downloads the attachment.

        You can use the 'path' parameter
        to save the downloaded information to a file.

        If you provide a path to a directory and not a file,
        then we will use the name of the remote file as the name.

        :param path: Path to save the data to, optional
        :type path: str
        :return: Downloaded bytes
        :rtype: bytes
        """

        # Get our paths:

        path = self._create_name(path)

        # Call the 'low_download' function with our URL:

        return self.low_download(self.url, path=path)

    def download_thumbnail(self, path: str=None) -> bytes:
        """
        Downloads the thumbnail of this attachment.

        You can use the 'path' parameter 
        to save the thumbnail to a file.

        Like the download method,
        we automaticallygenerate a name if the path is None,
        or is just a directory.

        :param path: Path to save the file to, defaults to None
        :type path: str, optional
        :return: downloaded bytes
        :rtype: bytes
        """

        # Get our path:

        path = self._create_name(path)

        # Call the 'low_download' function with our URL:

        return self.low_download(self.url, path=path)

    def _create_name(self, name: str=None) -> str:
        """
        Creates a name for the download file.

        If the name is none, then we return the
        title of this addon.

        We return this name as a path.

        :param name: Name to use
        :type name: str
        :return: Path to save the file under
        :rtype: str
        """

        if name is not None:

            # Check if we are working with a directory:

            if isdir(name):

                # Get the name of the file:

                return self.url.split('/')[-1]

            else:

                # Just use the name provided:

                return name

        # Otherwise, return the None:

        return None


@dataclass
class CurseFile(BaseDownloader):

    """
    CurseFile - Represents an addon file.

    An 'addon file' is a specific file available in an addon.
    Because multiple versions for an addon can exist,
    most addons have multiple files that can be downloaded.

    We contain useful metadata on each file,
    and offer the ability to download said file,
    as well as retrieve the changelog.

    We contain the following parameters:

        * id - ID of the file
        * addon_id - ID of the addon this file is apart of
        * display_name - Display name of the file
        * file_name - File name of the file
        * date - Date the file was uploaded
        * download_url - Download URL of the file
        * length - Length in bytes of the file
        * version - Version of the game needed to work with this file
        * dependencies - List dependency addons ID's
    """

    id: int
    addon_id: int
    display_name: str
    file_name: str
    date: str
    download_url: str
    length: int
    version: Tuple[str, ...]
    dependencies: Tuple[int, ...]

    INST_ID = 6

    @property
    def changelog(self) -> CurseDescription:
        """
        Gets the changelog of this file,
        and returns it as a CurseDescription object.

        Whatever handler is registered to the HandlerCollection for 
        getting game descriptions will be used!

        :return: CurseDescription representing the description of this file
        :rtype: CurseDescription
        """

        return self.hands.file_description(self.addon_id, self.id)

    def get_dependencies(self) -> Tuple[CurseAddon, ...]:
        """
        Returns a tuple of all CurseAddons
        that are the dependencies of this file.

        :return: Tuple of CurseAddons
        :rtype: Tuple[CurseAddon]
        """

        # Iterate over the dependency IDs:

        final = []

        for depen_id in self.dependencies:

            # Get and create the CurseAddon instances:

            final.append(self.hands.addon(depen_id))

        # Return the final tuple:

        return tuple(final)

    def get_addon(self) -> CurseAddon:
        """
        Gets the CurseAddon this file is apart of.

        :return: CurseAddon this file is apart of
        :rtype: CurseAddon
        """

        return self.hands.addon(self.addon_id)

    def download(self, path: str=None) -> bytes:
        """
        Downloads the file.

        If the provided path points to a directory,
        then the default file name will be used.

        :param path: Path to download the file to, optional
        :type path: str
        :return: Downloaded file bytes
        :rtype: bytes
        """

        # See if we should do some path handling:

        temp_path = None

        if path is not None:

            # Check if we are working with a directory:

            if isdir(path):

                # Join the paths:

                temp_path = join(path, self.file_name)

            else:

                # Just set the target to what was given:

                temp_path = path

        # Do the download operation:

        return self.low_download(self.download_url, path=temp_path)


@dataclass
class CurseAddon(BaseCurseInstance):

    """
    CurseAddon - Represents a addon for a specific game.

    An 'addon' is something(mod, modpack, skin) that is added onto a game
    to change or add certain features and aspects.

    The definition of a generic addon is purposefly left ambiguous,
    as it can be anything from a resource pack in minecraft to a mod in Kerbal Space Program.
    If you want something a bit more helpful and specific to a game,
    then the game wrappers might be helpful!

    This class contains useful metadata on a addon.
    We do NOT contain a detailed description information.

    We contain the following parameters:

        * name - Name of the addon
        * slug - Slug of the addon
        * summary - Short summary of the addon
        * url - URL of the addon page
        * lang - Language of the addon
        * date_created - Date of creation
        * date_modified - Date of last modification
        * date_release - Date of the latest release
        * id - ID of this addon
        * download_count - Number of downloads
        * game_id - ID of the game this addon is in
        * available - Boolean determining if the addon is available
        * experimental - Boolean determining if the addon is expiremental
        * authors - Tuple of authors for this addon
        * attachments - Tuple of attachments attributed with this addon
        * category_id - ID of the category this addon is in
        * is_featured - Boolean determining if this addon is featured
        * popularity_score - Float representing this addon's popularity score
            (Most likely used for popularity ranking)
        * popularity_rank - Int representing the game's popularity rank
        * game_name - Name of the game
    """

    name: str
    slug: str
    summary: str
    url: str
    lang: str
    date_created: str
    date_modified: str
    date_release: str
    id: int
    download_count: int
    game_id: int
    available: bool
    experimental: bool
    authors: Tuple[CurseAuthor, ...]
    attachments: Tuple[CurseAttachment, ...]
    category_id: int
    is_featured: bool
    popularity_score: int
    popularity_rank: int
    game_name: str

    INS_ID = 5

    @property
    def description(self) -> CurseDescription:
        """
        Gets the description for this CurseAddon.

        Whatever handler is assigned to addon descriptions will be
        used to generate a CurseDescription.

        :return: CurseDescription representing this addon
        :rtype: CurseDescription
        """

        return self.hands.addon_description(self.id)

    def files(self) -> Tuple[CurseFile, ...]:
        """
        Gets a tuple of files associated with this addon.

        :return: Tuple of CurseFile instances
        :rtype: Tuple[CurseFile, ...]
        """
        
        return self.hands.addon_files(self.id)

    def file(self, file_id: int) -> CurseFile:
        """
        Gets a file associated with this addon.
        We use this addon ID to get the file.

        :param file_id: ID of the file to find
        :type file_id: int
        :return: CurseFile representing the file
        :rtype: CurseFile
        """

        return self.hands.addon_file(self.id, file_id)

    def game(self) -> CurseGame:
        """
        Gets and returns the CurseGame instance
        this addon is apart of.

        :return: CurseGame instance
        :rtype: CurseGame
        """

        return self.hands.game(self.game_id)

    def category(self) -> CurseCategory:
        """
        Gets and returns the category we are a member of

        :return: CurseCategory of the category we are in
        :rtype: CurseCategory
        """

        # Get and return the category:

        return self.hands.category(self.category_id)


@dataclass
class CurseCategory(BaseCurseInstance):

    """
    CurseCategory - Represents a category for a specific game. 

    A 'category' is a collection of addons for a specific game.
    Lets use Minecraft as an example, Minecraft has the following catagories:

        * Resource Packs
        * Modpacks
        * Mods
        * Worlds

    If you wanted to get info on a mod, then you would search in the 'Mods' category.
    This allows addons to be organized by type.

    This class contains identifying info on catagories,
    most important being the category ID,
    which is needed to search for projects in that specific category.

    Categories can be nested,
    meaning that each category can have a variable number of child categories.
    We contain references to the root_id, 
    which is the ID of the root category for all nested categories.
    The parent_id is the ID of the category one level up from ours.

    We contain the following parameters:

        * id - ID of the category
        * game_id - ID of the game this category is under
        * name - Name of the category
        * root_id - ID of the root category
        * parent_id - ID of the parent category
        * icon - CurseAttachment of this catagories icon
        * date - Date this category was created 
    """

    id: int
    game_id: int
    name: str
    root_id: int
    parent_id: int
    icon: CurseAttachment
    date: str

    INST_ID = 1

    def sub_categories(self) -> Tuple[CurseAddon, ...]:
        """
        Gets all sub-catagories for this category.

        :return: Tuple of sub-categories for this category
        :rtype: Tuple[CurseCategory, ...]
        """

        return self.hands.sub_category(self.id)

    def parent_category(self) -> CurseCategory:
        """
        Gets and returns the parent category for this instance.

        If there is no parent category, 
        then 'None' will be returned instead

        :return: CurseCategory instance, or None if there is no parent
        :rtype: CurseCategory, None
        """

        # Check if we even have a parent:

        if self.parent_category is None:

            # No parent, return None:

            return None

        # Get the parent category:

        return self.hands.category(self.parent_id)

    def root_category(self) -> CurseCategory:
        """
        Gets and returns the root category for this instance.

        If there is no parent category,
        then 'None' will be returned instead.

        :return: CurseCategory instance, or None if there is no parent
        :rtype: CurseCategory, None
        """

        # Check if we even have a root category:

        if self.parent_category is None:

            # No parent, return None:

            return None

        return self.hands.category(self.root_id)

    def search(self, search_param: Optional[SearchParam]=None) -> Tuple[CurseAddon]:
        """
        Searches this category with the given search parameters.

        :param search_param: Search parameter object to use, defaults to None
        :type search_param: Optional[BaseSearch], optional
        :return: Tuple of curse addons found in the search
        :rtype: Tuple[CurseAddon, ...]
        """

        return self.hands.search(self.game_id, self.id, search_param)

    def iter_search(self, search_param: Optional[SearchParam]=None) -> CurseAddon:
        """
        Invokes the 'iter_search' method of the HC with the given search parameters.

        You should check out HC's documentation on 'iter_search',
        but in a nutshell it basically allows you to iterate
        though all found addons, automatically incrementing the index when necessary.

        :param search_param: [description], defaults to None
        :type search_param: Optional[SearchParam], optional
        :return: Each curse addon found
        :rtype: CurseAddon
        """

        # Return the results from 'iter_search':

        return self.hands.iter_search(self.game_id, self.id, search_param)


@dataclass
class CurseGame(BaseCurseInstance):

    """
    CurseGame - Represents a game on curseforge.

    A game is a collection of addons/catagories
    that are relevant to a specific title, 
    such as World of Warcraft or Minecraft.

    Game instances can have useful meta info for a particular title,
    but the most important are the game ID and catagories info,
    as you will need these values to get more precise project information.

    Not all games support addons!
    Some games are listed for reasons unknown,
    and do not have addon support.
    You can determine if a game has addon support by checking the 'support_addons' parameter.

    Some values are not loaded and interpreted by this class!
    These include game detection hints(Used for determining where a game is),
    and game file information(What files are necessary for modding and where they can be found).
    If this information is important to you, then the raw data can be found under the 'raw' parameter. 

    We contain the following parameters:

        * name - Name of the game
        * slug - Slug of the game
        * id - ID of the game
        * support_addons - Boolean determining if this game supports addons
        * cat_ids - ID's of the root categories associated with this game,
            use the 'categories' method to get CategoryInstances for these ID's
    """

    name: str
    slug: str
    id: int
    support_addons: float
    cat_ids: Tuple[int,...]

    INST_ID = 2

    def categories(self) -> Tuple[CurseCategory, ...]:
        """
        Gets a tuple of all root categories for this game.

        :return: Tuple of all root CurseCategories
        :rtype: Tuple[CurseCategory, ...]
        """

        # Get all categories for this game:

        final = []

        for cat_id in self.cat_ids:

            # Retrieve category info for this ID:

            final.append(self.hands.category(cat_id))

        # Return the final tuple:

        return tuple(final)
